#! /usr/bin/perl -w

# rbl-lookup

# Scan an email to see if it has come from/via a machine listed
# in any of the MAPS lists and/or most of the ORBS lists.

# Copyright (C) 2000, 2001 Darren Salt <ds@youmustbejoking.demon.co.uk>

# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You may have received a copy of the GNU General Public License along with
# this program (it's likely that you'll have a copy elsewhere); if not,
# write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  US

$PROG = $0;
$PROG =~ s%^.*/%%;

$version = $PROG.' 1.4.0, 17 Jan 2002

Copyright (C) 2000-2002 Darren Salt.
This is free software; see the source for copying conditions. There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.';

$etc = '/home/ds';

# Prototypes
sub do_lookup ($@);
sub make_list ($%);
sub herror ($$);
sub verbose (;$);
sub make_label (@);

# Domains used in lookups
@maps = ('.mail-abuse.org', '.maps.vix.com');
@ordb = ('.ordb.org');
@orbl = ('.orbl.org');
@orbz = ('.orbz.org');

$help = 'Usage: '.$PROG.' [options] [hosts|addresses]
Reads one email from stdin, reads the Received headers and writes
the email to stdout (possibly with X-RBL-Warning headers).
Alternatively, reports on addresses given on the command line.

   -i --ignore-first=...	ignore the first n Received headers
   -m --maps=...		lookup IPs via *'.$maps[0].'
		You can restrict the lookups as follows ("none"=ignore):
		B = blackholes		(blacklisted IPs)
		R = relays		(known open relays)
		D = dialups		(abused dialups)
		The default is "BR".
		See http://www'.$maps[0].'/ for details.
   -M --vix			use '.$maps[1].' instead of '.$maps[0].'
      --ordb=...		lookup IPs via *'.$ordb[0].'
		R = relays		(reported/known? open relays)
		Default is "R".
		See http://www.'.$ordb[0].'/ for details.
      --orbl=...		lookup IPs via *'.$orbl[0].'
		R = or		(reported/known? open relays)
		Default is "R".
		See http://www.'.$orbl[0].'/ for details.
   -o --orbz=...
      --orbs=...
		I = inputs		(reported open inputs)
		O = outputs		(reported open outputs)
		Default is "IO".
		See http://www.'.$orbz[0].'/ for details.
      --all=...
		Set the default lookup action: all, std, none.
		Without this option, the default is "std".
		With this option, the default is "all".
   -f --full			lookup IP addresses *and* named hosts
   -r --report			report only (don\'t write out the email)
   -v --verbose			be verbose (more -v = more verbose)
      --help			this help
      --version			print version
';

use Socket;
use Getopt::Long;
use vars qw($opt_vix $opt_test $opt_full $opt_debug);

Getopt::Long::Configure ('bundling');
GetOptions ('ignore-first|i:s', 'maps|m:s', 'vix|M', 'ordb:s', 'orbl:s', 'orbz|orbs|o:s',
	    'all:s', 'full|f', 'report|r', 'verbose|v+', 'test', 'help|?', 'version',
	    'debug|d')
  or die "Try '$0 --help'.\n";

print $version, "\n" if defined $opt_help || defined $opt_version;
print "\n", $help if defined $opt_help;
exit 0 if defined $opt_help || defined $opt_version;

$| = 1;

sub set_std($$$)
{
  my ($opt, $all, $std) = @_;
  return $all if $opt eq 'all';
  return $std if $opt eq 'std';
  return ' ' if $opt eq 'none';
  $opt =~ tr/a-z/A-Z/;
  my $sub = '(?-xism:[^'.$all.'])';
  $opt =~ s/$sub//g;
  return $opt;
}

$opt_ignore_first = 1
  if defined $opt_ignore_first && $opt_ignore_first !~ /^[[:digit:]]+$/;
$opt_ignore_first = 0 unless defined $opt_ignore_first;
$opt_all = 'all' if defined $opt_all && $opt_all eq '';
$opt_all = 'std' unless defined $opt_all;
$opt_maps = $opt_all unless defined $opt_maps;
$opt_ordb = $opt_all unless defined $opt_ordb;
$opt_orbl = $opt_all unless defined $opt_orbl;
$opt_orbz = $opt_all unless defined $opt_orbz;

$opt_maps = set_std ($opt_maps, 'BRD', 'BR');
$opt_ordb = set_std ($opt_ordb, 'R', 'R');
$opt_orbl = set_std ($opt_orbl, 'R', 'R');
$opt_orbz = set_std ($opt_orbz, 'IO', 'IO');

# IP addresses which may be returned by lookups
for $i (0..9) {
  push @ip, inet_aton ('127.0.0.'.$i);
}

# Read in some configuration...
if (-f ($etc.'/rbl-lookup.conf')) {
  open CONFIG,'< '.$etc.'/rbl-lookup.conf' or die $!;
  local $line;
  while ($line = <CONFIG>)
  {
    chomp $line;
    $line =~ s/^[[:space:]]+//; # leading spaces
    $line =~ s/[[:space:]]*(?:#.*)?$//; # trailing spaces, comment
    next if $line eq '';
    if ($line =~ /^trusted-domain\b[[:space:]]*(.*)/io) {
      push @trusted_dom, split (/[[:space:]]+/o, $1);
    } elsif ($line =~ /^trusted-ip\b[[:space:]]*(.*)/io) {
      push @trusted_ip, split (/[[:space:]]+/o, $1);
    }
    # elsif ...
  }
  close CONFIG;
  local $i;
  foreach $i (@trusted_ip) {
    $i =~ tr/A-Z/a-z/;
    $i =~ /^([^\/]*)(?:\/(.*))?$/o;
    local ($ip, $mask) = ($1, defined $2 ? $2 : '255.255.255.255');
    $i = [inet_aton ($ip)];
    if ($mask =~ /^[[:digit:]]+$/) {
      $mask = ~ ((1 << (32 - $mask)) - 1);
      push @{$i},
	pack ('I4', (($mask >> 24) & 0xFF) | (($mask >> 8) & 0xFF00) |
		    (($mask << 8) & 0xFF0000) | (($mask << 24) & 0xFF000000));
    } else {
      push @{$i}, inet_aton ($mask);
    }
  }
}


@scans = ();
# each element is [dns host] or [dns host, returned IP]

if (defined $opt_vix) {
  push @scans, ['rbl'.$maps[1]] 	if $opt_maps =~ /B/;
  push @scans, ['rss'.$maps[1]] 	if $opt_maps =~ /R/;
  push @scans, ['dul'.$maps[1]] 	if $opt_maps =~ /D/;
} else {
  push @scans, ['blackholes'.$maps[0]]	if $opt_maps =~ /B/;
  push @scans, ['relays'.$maps[0]]	if $opt_maps =~ /R/;
  push @scans, ['dialups'.$maps[0]]	if $opt_maps =~ /D/;
}

push @scans, ['relays'.$ordb[0]]	if $opt_ordb =~ /R/;

push @scans, ['or'.$orbl[0]]		if $opt_orbl =~ /R/;

push @scans, ['inputs'.$orbz[0]]	if $opt_orbz =~ /I/;
push @scans, ['outputs'.$orbz[0]]	if $opt_orbz =~ /O/;

if (verbose (1)) {
  print STDERR "Using: $opt_maps $opt_ordb $opt_orbl $opt_orbz\n";
  for $i (@scans) {
    print "\n  ", ${$i}[0];
    print '/', inet_ntoa (${$i}[1]) if $#{$i} > 0;
  }
  print "\n\n";
}

push @ipb, [ inet_aton ('0.0.0.0'),	inet_aton ('255.0.0.0') ];	# old broadcast
push @ipb, [ inet_aton ('10.0.0.0'),	inet_aton ('255.0.0.0') ];	# RFC1918 /8 private
push @ipb, [ inet_aton ('127.0.0.0'),	inet_aton ('255.0.0.0') ];	# localhost
push @ipb, [ inet_aton ('169.254.0.0'),	inet_aton ('255.255.0.0') ];	# local nets link
push @ipb, [ inet_aton ('172.16.0.0'),	inet_aton ('255.240.0.0') ];	# RFC1918 /12 private
push @ipb, [ inet_aton ('192.0.2.0'),	inet_aton ('255.255.255.0') ];	# network test
push @ipb, [ inet_aton ('192.168.0.0'),	inet_aton ('255.255.0.0') ];	# RFC1918 /16 private
push @ipb, [ inet_aton ('224.0.0.0'),	inet_aton ('240.0.0.0') ];	# multicast
push @ipb, [ inet_aton ('240.0.0.0'),	inet_aton ('248.0.0.0') ];	# reserved
push @ipb, [ inet_aton ('248.0.0.0'),	inet_aton ('248.0.0.0') ];	# unallocated
push @ipb, [ inet_aton ('255.255.255.255'), inet_aton ('255.255.255.255') ]; # broadcast

$headers = '';
$body = '';
@header = ();

if ($#ARGV >= 0) {
  # Reading from the CLI - fake some Received headers. This implies --report
  # - rewrite avoidance :-)
  foreach $i (@ARGV) {
    local $try;
    local @block;
    $i .= '.' unless substr ($i, -1) eq '.';
    foreach $try (1..4) { # four tries...
      ($_, $_, $_, $_, @block) = gethostbyname ($i.'.');
      last unless $? == 2;
    }
    herror ($?, $i) if $? > 0;
    foreach $try (@block) {
      push @header, 'Received: from you (' . inet_ntoa ($try) . ') by me';
    }
  }
  $opt_report = 1;
} else {
  # Read the headers from stdin
  while ($line = <>) {
    last if $line =~ /^\r?\n?$/o;
    $headers .= $line;
    chomp $line;
    $line =~ tr/\t/ /;
    $line =~ s/  +/ /go;
    $line =~ s/ +$//o;
    if ($line =~ /^[ \t]/) {
      $header[$#header] .= $line;
    } else {
      push @header, $line;
    }
  }
  # Remove any existing warnings
  $headers =~ s/^X-RBL-Warning:[ \t][^\n]*\n([ \t][^\n]*\n)*//imgo;
  # Slurp in the message body
  undef $/;
  $body = <>;
  $body = '' unless defined $body;
}

# Examine each Received header for machine names or IP addresses
foreach $i (@header) {
  next unless $i =~ /^Received: /io;
  next if --$opt_ignore_first > 0;
  # Find the rightmost name/addr in the 'from' section
  $i =~ /^(?i:Received): from ([^ ]+) (?:\(([^ )]+)(?: ([^ )]+))?)?/io;
  next unless defined $1;
  local @res = ($1, $2, $3);
  undef $res[2] if defined $res[2] && $res[2] =~ /^(he|eh)lo=/;
  undef $res[1] if defined $res[1] && $res[1] =~ /^(he|eh)lo=/;
  local $name = defined $res[2] ? $res[2] : defined $res[1] ? $res[1] : $res[0];
  next if $name =~ /^(he|eh)lo=/;
  # Strip out any surrounding []
  if ($name =~ /^\[(.*)]$/o) {
    $name = $1;
  } elsif (defined $opt_full) {
    print 'Ignoring ', $name, "\n" if verbose (3);
    next;
  }
  # Translate to, or lookup, network address
  local $ip = inet_aton ($name);
  next unless defined $ip;
  # Ignore certain netblocks
  next if 0 < grep { ($ip & ${$_}[1]) eq ${$_}[0] } @ipb;
  # Ignore if in the trusted IPs list
  next if defined @trusted_ip &&
	  0 < grep { ($ip & ${$_}[1]) eq ${$_}[0] } @trusted_ip;
  # try to get the canonical name
  foreach $try (1..4) { # four tries (allowing for things like dialup)...
    ($name2) = gethostbyaddr ($ip, AF_INET);
    last unless $? == 2;
  }
  if ($? > 1) {
    herror ($?, $name);
  } else {
    $name2 = $name unless defined $name2;
    $name2 =~ tr/A-Z/a-z/;
    next if 0 < grep { substr ($name2, -length ($_)) eq $_ } @trusted_dom;
    $lookup{$ip} = $name2;
  }
}

if (defined %lookup) {
  # Do the lookups
  foreach $i (keys %lookup) {
    foreach $j (@scans) {
      do_lookup ($i, @{$j});
    }
  }
  # Wait for the lookups to be completed
  while ($child = wait) {
    local $ret = $? / 256;
    last if $child == -1;
    $found{${$children{$child}}[0]}{${$children{$child}}[1]} = 1 if $ret == 1;
    undef $children{$child};
  }
}

# Write the message to stdout, possibly with an X-RBL-Warning header
(print $headers or die $!) unless defined $opt_report;
#$rbl = '';
foreach $i (@scans) {
  my $label = make_label (@{$i});
  print make_list ($label, %{$found{$label}}) or die $!;
}
#(print $rbl or die $!) if $rbl ne '';
(print "\n", $body or die $!) unless defined $opt_report || defined $opt_debug;
exit 0;


sub make_label (@)
{
  return $_[0].($#_ > 0 ? ('/'.inet_ntoa ($_[1])) : '');
}


sub do_lookup ($@)
# Look up the given address in the given list ** IPv4 **
# (try to fork it)
{
  local ($ip, @type) = @_;
  local $label = make_label (@type);
  local $iptype = $ip.'.'.$label;
  return if defined $lookups{$iptype};
  # Right. We haven't tried to look this one up before...
  $lookups{$iptype} = '';
  $lookup{$ip} = inet_ntoa ($ip) unless defined $lookup{$ip};
  # Try to fork
  local $child = defined $opt_debug ? 0 : fork;
  if ($child > 0) {
    # Parent (successful fork)
    $children{$child} = [$label, $ip];
  } else {
    # Child (successful fork) OR failed fork
    # - either way, we want to perform the lookup
    # Convert the address into dotted quad
    local @quad = unpack ('C4', $ip);
    print STDERR "Looking up $lookup{$ip} via $label\n" if verbose (3);
    # ... reverse it, append the domain and look it up
    local @block;
    local $try;
    local $name = $quad[3].'.'.$quad[2].'.'.$quad[1].'.'.$quad[0].'.'.$type[0];
    foreach $try (1..4) { # four tries...
      ($_, $_, $_, $_, @block) = gethostbyname ($name);
      last unless $? == 2;
    }
    herror ($?, $lookup{$ip}.' ('.$label.')') if $? > 0;
    my @got;
    if ($#type > 0) {
      @got = grep {$_ = $type[1]} @block;
    } else {
      @got = @block;
    }
    print STDERR '* Found ', $lookup{$ip}, ' (', $label, ")\n"
      if $#got >= 0 && verbose;
    # If this is a child process, exit with 0 = clear (or timed out) or 1 = is listed
    exit ($#got >= 0) ? 1 : 0 if $child == 0 && not defined $opt_debug;
    # Otherwise record the address/list pair and just return
    $found{$label}{$ip} = 1 if $#got >= 0;
  }
}


sub make_list ($%)
# Build a list of X-RBL-Warning headers from the listed addresses
{
  my ($head, %list) = @_;
  my $text = '';
  my @keys = sort keys %list;
  return '' if $#keys < 0;
  if (defined $opt_report) {
    $text .= $head.":\n";
    foreach my $i (@keys) {
      $text .= "\t".$lookup{$i}."\n";
    }
  } else {
    foreach my $i (@keys) {
      $text .= 'X-RBL-Warning: '.$head.': '.$lookup{$i}."\n" if $list{$i};
    }
  }
  return $text;
}


sub herror ($$)
# Report a lookup error...
{
  return unless verbose (1);
  local ($err, $name) = @_;
  if	($err == 1) { print '! ', $name, " not found (authoritative)\n" if verbose (2); }
  elsif ($err == 2) { print '! ', $name, " not found (non-authoritative)\n" if verbose (2); }
  elsif ($err == 3) { print '! ', $name, ": non-recoverable lookup error\n" if verbose; }
  elsif ($err == 4) { print '! ', $name, " exists, but has no A record\n" if verbose (2); }
  else		    { print '! ', $name, ": unknown lookup error\n" if verbose; }
}


sub verbose (;$)
{
  local $i = shift;
  return (defined $opt_verbose && $opt_verbose >= (defined $i? $i : 1))
    || (defined $opt_report && not defined $i);
}
